local AceLocale = LibStub("AceLocale-3.0")
local L = AceLocale:GetLocale("Spy")

function Spy:RefreshCurrentList(player, source)
	local MainWindow = Spy.MainWindow
	if not MainWindow:IsShown() then
		return
	end

	local mode = Spy.db.profile.CurrentList
	local manageFunction = Spy.ListTypes[mode][2]
	if manageFunction then manageFunction() end

	local button = 1
	for index, data in pairs(Spy.CurrentList) do
		if button <= Spy.ButtonLimit then
			local level = "??"
			local class = "UNKNOWN"
			local opacity = 1

			local playerData = SpyPerCharDB.PlayerData[data.player]
			if playerData then
				if playerData.level then level = playerData.level end
				if playerData.class then class = playerData.class end
			end
			if mode == 1 and Spy.InactiveList[data.player] then
				opacity = 0.5
			end

			if player == data.player then
				if not source or source ~= UnitName("player") then
					Spy:AlertPlayer(player)
					if not source then Spy:AnnouncePlayer(player) end
				end
			end

			Spy:SetBar(button, data.player, level.." "..L[class], 100, "Class", class, nil, opacity)
			Spy.ButtonName[button] = data.player
			button = button + 1
		end
	end
	Spy.ListAmountDisplayed = button - 1

	if Spy.db.profile.ResizeSpy then
		Spy:AutomaticallyResize()
	end

	Spy:ManageBarsDisplayed()
end

function Spy:ManageNearbyList()
	local prioritiseKoS = Spy.db.profile.PrioritiseKoS

	local activeKoS = {}
	local active = {}
	for player in pairs(Spy.ActiveList) do
		if prioritiseKoS and SpyPerCharDB.KOSData[player] then
			table.insert(activeKoS, { player = player, time = Spy.NearbyList[player] })
		else
			table.insert(active, { player = player, time = Spy.NearbyList[player] })
		end
	end

	local inactiveKoS = {}
	local inactive = {}
	for player in pairs(Spy.InactiveList) do
		if prioritiseKoS and SpyPerCharDB.KOSData[player] then
			table.insert(inactiveKoS, { player = player, time = Spy.NearbyList[player] })
		else
			table.insert(inactive, { player = player, time = Spy.NearbyList[player] })
		end
	end

	table.sort(activeKoS, function(a, b) return a.time < b.time end)
	table.sort(inactiveKoS, function(a, b) return a.time < b.time end)
	table.sort(active, function(a, b) return a.time < b.time end)
	table.sort(inactive, function(a, b) return a.time < b.time end)

	local list = {}
	for player in pairs(activeKoS) do table.insert(list, activeKoS[player]) end
	for player in pairs(inactiveKoS) do table.insert(list, inactiveKoS[player]) end
	for player in pairs(active) do table.insert(list, active[player]) end
	for player in pairs(inactive) do table.insert(list, inactive[player]) end
	Spy.CurrentList = list
end

function Spy:ManageLastHourList()
	local list = {}
	for player in pairs(Spy.LastHourList) do
		table.insert(list, { player = player, time = Spy.LastHourList[player] })
	end
	table.sort(list, function(a, b) return a.time > b.time end)
	Spy.CurrentList = list
end

function Spy:ManageIgnoreList()
	local list = {}
	for player in pairs(SpyPerCharDB.IgnoreData) do
		table.insert(list, { player = player, time = SpyPerCharDB.PlayerData[player].time })
	end
	table.sort(list, function(a, b) return a.time > b.time end)
	Spy.CurrentList = list
end

function Spy:ManageKillOnSightList()
	local list = {}
	for player in pairs(SpyPerCharDB.KOSData) do
		table.insert(list, { player = player, time = SpyPerCharDB.PlayerData[player].time })
	end
	table.sort(list, function(a, b) return a.time > b.time end)
	Spy.CurrentList = list
end

function Spy:ManageExpirations()
	local mode = Spy.db.profile.CurrentList
	local expirationFunction = Spy.ListTypes[mode][3]
	if expirationFunction then expirationFunction() end
end

function Spy:ManageNearbyListExpirations()
	local expired = false
	local currentTime = time()
	for player in pairs(Spy.ActiveList) do
		if (currentTime - Spy.ActiveList[player]) > Spy.ActiveTimeout then
			Spy.InactiveList[player] = Spy.ActiveList[player]
			Spy.ActiveList[player] = nil
			expired = true
		end
	end
	if Spy.db.profile.RemoveUndetected ~= "Never" then
		for player in pairs(Spy.InactiveList) do
			if (currentTime - Spy.InactiveList[player]) > Spy.InactiveTimeout then
				Spy.InactiveList[player] = nil
				Spy.NearbyList[player] = nil
				expired = true
			end
		end
	end
	if expired then
		Spy:RefreshCurrentList()
		if Spy.db.profile.HideSpy then
			local entries = 0
			for _ in pairs(Spy.NearbyList) do
				entries = entries + 1
			end
			if entries == 0 then
				Spy.MainWindow:Hide()
			end
		end
	end
end

function Spy:ManageLastHourListExpirations()
	local expired = false
	local currentTime = time()
	for player in pairs(Spy.LastHourList) do
		if (currentTime - Spy.LastHourList[player]) > 3600 then
			Spy.LastHourList[player] = nil
			expired = true
		end
	end
	if expired then
		Spy:RefreshCurrentList()
	end
end

function Spy:RemovePlayerFromList(player)
	Spy.NearbyList[player] = nil
	Spy.ActiveList[player] = nil
	Spy.InactiveList[player] = nil
	Spy:RefreshCurrentList()
end

function Spy:ClearList()
	Spy.NearbyList = {}
	Spy.ActiveList = {}
	Spy.InactiveList = {}
	Spy.ListAmountDisplayed = 0
	Spy:SetCurrentList(1)
	if IsControlKeyDown() then
		Spy:EnableSpy(not Spy.db.profile.Enabled, false)
	end
end

function Spy:AddPlayerData(name, class, level, race, guild, isEnemy, isGuess)
	local info = {}
	info.class = class
	info.level = level
	info.race = race
	info.guild = guild
	info.isEnemy = isEnemy
	info.isGuess = isGuess
	SpyPerCharDB.PlayerData[name] = info
	return SpyPerCharDB.PlayerData[name]
end

function Spy:UpdatePlayerData(name, class, level, race, guild, isEnemy, isGuess)
	local detected = true
	local playerData = SpyPerCharDB.PlayerData[name]
	if not playerData then
		playerData = Spy:AddPlayerData(name, class, level, race, guild, isEnemy, isGuess)
	else
		if class then playerData.class = class end
		if level then playerData.level = level end
		if race then playerData.race = race end
		if guild then playerData.guild = guild end
		if isEnemy then playerData.isEnemy = isEnemy end
		if isGuess then playerData.isGuess = isGuess end
	end

	if WorldMapFrame:IsVisible() then
		SetMapToCurrentZone()
	end
	local mapX, mapY = GetPlayerMapPosition("player")
	if playerData and mapX ~= 0 and mapY ~= 0 then
		mapX = math.floor(tonumber(mapX) * 100) / 100
		mapY = math.floor(tonumber(mapY) * 100) / 100

		playerData.time = time()
		playerData.zone = GetZoneText()
		playerData.subZone = GetSubZoneText()
		playerData.mapX = mapX
		playerData.mapY = mapY
	else
		detected = false
	end
	return detected
end

function Spy:RemovePlayerData(name)
	SpyPerCharDB.PlayerData[name] = nil
end

function Spy:AddIgnoreData(name)
	SpyPerCharDB.IgnoreData[name] = true
end

function Spy:RemoveIgnoreData(name)
	if SpyPerCharDB.IgnoreData[name] then
		SpyPerCharDB.IgnoreData[name] = nil
	end
end

function Spy:AddKOSData(name)
	SpyPerCharDB.KOSData[name] = time()
end

function Spy:RemoveKOSData(name)
	if SpyPerCharDB.KOSData[name] then
		SpyPerCharDB.KOSData[name] = nil
	end
end

function Spy:AlertPlayer(player)
	if SpyPerCharDB.KOSData[player] then
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\detected-kos.wav")
		end
		if Spy.db.profile.WarnOnKOS then
			local text = Spy.db.profile.Colors.Warning["Warning Text"]
			local msg = L["KoSWarning"]..player
			UIErrorsFrame:AddMessage(msg, text.r, text.g, text.b, 1.0, UIERRORS_HOLD_TIME)
		end
	else
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\detected-nearby.wav")
		end
	end
end

function Spy:AlertStealthPlayer(player)
	if Spy.db.profile.EnableSound then
		PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\detected-stealth.wav")
	end
	if Spy.db.profile.WarnOnStealth then
		local text = Spy.db.profile.Colors.Warning["Warning Text"]
		local msg = L["StealthWarning"]..player
		UIErrorsFrame:AddMessage(msg, text.r, text.g, text.b, 1.0, UIERRORS_HOLD_TIME)
	end
end

function Spy:AnnouncePlayer(player, channel)
	local msg = ""
	local isKOS = SpyPerCharDB.KOSData[player]
	local playerData = SpyPerCharDB.PlayerData[player]

	local announce = Spy.db.profile.Announce
	if channel or announce == "Self" or (announce == "Guild" and GetGuildInfo("player") ~= nil and not Spy.InInstance) or (announce == "Party" and GetNumPartyMembers() > 0) or (announce == "Raid" and UnitInRaid("player")) then
		if announce == "Self" and not channel then
			if isKOS then
				msg = msg..L["SpySignatureColored"]..L["KillOnSightDetectedColored"]..player.." "
			else
				msg = msg..L["SpySignatureColored"]..L["PlayerDetectedColored"]..player.." "
			end
		else
			if isKOS then
				msg = msg..L["KillOnSightDetected"]..player.." "
			else
				msg = msg..L["PlayerDetected"]..player.." "
			end
		end
		if playerData then
			if playerData.guild and playerData.guild ~= "" then
				msg = msg.."<"..playerData.guild.."> "
			end
			if playerData.level or playerData.race or (playerData.class and playerData.class ~= "") then
				msg = msg.."- "
				if playerData.level and playerData.level > 0 then msg = msg..L["Level"].." "..playerData.level.." " end
				if playerData.race and playerData.race ~= "" then msg = msg..playerData.race.." " end
				if playerData.class and playerData.class ~= "" then msg = msg..L[playerData.class].." " end
			end
			if playerData.zone then
				if playerData.subZone and playerData.subZone ~= "" and playerData.subZone ~= playerData.zone then
					msg = msg.."- "..playerData.subZone..", "..playerData.zone
				else
					msg = msg.."- "..playerData.zone
				end
			end
			if playerData.mapX and playerData.mapY then msg = msg.." ("..math.floor(tonumber(playerData.mapX) * 100)..","..math.floor(tonumber(playerData.mapY) * 100)..")" end
		end

		if channel then
			-- announce to selected channel
			if channel == "PARTY" or channel == "RAID" or channel == "GUILD" then
				SendChatMessage(msg, channel)
			elseif channel == "LOCAL" then
				SendChatMessage(msg, "CHANNEL", nil, GetChannelName(L["LocalDefenseChannelName"].." - "..GetZoneText()))
			end
		else
			-- announce to standard channel
			if isKOS or not Spy.db.profile.OnlyAnnounceKoS then
				if announce == "Self" then
					DEFAULT_CHAT_FRAME:AddMessage(msg)
				elseif announce == "LocalDefense" then
					SendChatMessage(msg, "CHANNEL", nil, GetChannelName(L["LocalDefenseChannelName"].." - "..GetZoneText()))
				else
					SendChatMessage(msg, strupper(announce))
				end
			end
		end
	end

	-- announce to other Spy users
	if Spy.db.profile.ShareData then
		local class, level, race, guild, zone, subZone, mapX, mapY = "", "", "", "", "", "", "", ""
		if playerData then
			if playerData.class then class = playerData.class end
			if playerData.level then level = playerData.level end
			if playerData.race then race = playerData.race end
			if playerData.guild then guild = playerData.guild end
			if playerData.zone then zone = playerData.zone end
			if playerData.subZone then subZone = playerData.subZone end
			if playerData.mapX then mapX = playerData.mapX end
			if playerData.mapY then mapY = playerData.mapY end
		end
		local details = Spy.Version.."|"..player.."|"..class.."|"..level.."|"..race.."|"..zone.."|"..subZone.."|"..mapX.."|"..mapY.."|"..guild
		if strlen(details) < 240 then
			if channel then
				if (channel == "PARTY" and GetNumPartyMembers() > 0) or (channel == "RAID" and UnitInRaid("player")) or (channel == "GUILD" and GetGuildInfo("player") ~= nil) then
					Spy:SendCommMessage(Spy.Signature, details, channel)
				end
			else
				if GetNumPartyMembers() > 0 then Spy:SendCommMessage(Spy.Signature, details, "PARTY") end
				if UnitInRaid("player") then Spy:SendCommMessage(Spy.Signature, details, "RAID") end
				if Spy.InInstance == false and GetGuildInfo("player") ~= nil then Spy:SendCommMessage(Spy.Signature, details, "GUILD") end
			end
		end
	end
end

function Spy:ToggleIgnorePlayer(ignore, player)
	if ignore then
		Spy:AddIgnoreData(player)
		Spy:RemoveKOSData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-add.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerAddedToIgnoreColored"]..player)
	else
		Spy:RemoveIgnoreData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-remove.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerRemovedFromIgnoreColored"]..player)
	end
	Spy:RefreshCurrentList()
end

function Spy:ToggleKOSPlayer(kos, player)
	if kos then
		Spy:AddKOSData(player)
		Spy:RemoveIgnoreData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-add.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerAddedToKOSColored"]..player)
	else
		Spy:RemoveKOSData(player)
		if Spy.db.profile.EnableSound then
			PlaySoundFile("Interface\\AddOns\\Spy\\Sounds\\list-remove.wav")
		end
		DEFAULT_CHAT_FRAME:AddMessage(L["SpySignatureColored"]..L["PlayerRemovedFromKOSColored"]..player)
	end
	Spy:RefreshCurrentList()
end

function Spy:ButtonClicked()
	local name = Spy.ButtonName[this.id]
	if name and name ~= "" then
		if arg1 == "LeftButton" then
			if IsShiftKeyDown() then
				if SpyPerCharDB.KOSData[name] then
					Spy:ToggleKOSPlayer(false, name)
				else
					Spy:ToggleKOSPlayer(true, name)
				end
			elseif IsControlKeyDown() then
				if SpyPerCharDB.IgnoreData[name] then
					Spy:ToggleIgnorePlayer(false, name)
				else
					Spy:ToggleIgnorePlayer(true, name)
				end
			else
				this:SetAttribute("macrotext", "/targetexact "..name)
			end
		elseif arg1 == "RightButton" then
			Spy:BarDropDownOpen(this)
			CloseDropDownMenus(1)
			ToggleDropDownMenu(1, nil, Spy_BarDropDownMenu)
		end
	end
end

function Spy:ParseMinimapTooltip(tooltip)
	local newTooltip = ""
	local newLine = false
	for text in string.gmatch(tooltip, "[^\n]*") do
		local name = text
		if string.len(text) > 0 then
			if strsub(text, 1, 2) == "|c" then
				name = strsub(text, 11, -3)
			end
			local playerData = SpyPerCharDB.PlayerData[name]
			if not playerData then
				for index, _ in pairs(Spy.LastHourList) do
					local realmSeparator = strfind(index, "-")
					if realmSeparator and realmSeparator > 1 and strsub(index, 1, realmSeparator - 1) == strsub(name, 1, realmSeparator - 1) then
						playerData = SpyPerCharDB.PlayerData[index]
						break
					end
				end
			end
			if playerData and playerData.isEnemy then
				local desc = ""
				if playerData.class and playerData.level then
					desc = L["MinimapClassText"..playerData.class].."["..playerData.level.." "..L[playerData.class].."]|r"
				elseif playerData.class then
					desc = L["MinimapClassText"..playerData.class].."["..L[playerData.class].."]|r"
				elseif playerData.level then
					desc = "["..playerData.level.."]|r"
				end
				newTooltip = newTooltip..text.."|r "..desc
				if not SpyPerCharDB.IgnoreData[name] and not Spy.InInstance then
					local detected = Spy:UpdatePlayerData(name, nil, nil, nil, nil, true, nil)
					if detected then
						Spy:AddDetected(name, time(), false)
					end
				end
			else
				newTooltip = newTooltip..text.."|r"
			end
			newLine = false
		elseif not newLine then
			newTooltip = newTooltip.."\n"
			newLine = true
		end
	end
	return newTooltip
end

function Spy:ParseUnitAbility(analyseSpell, event, player, flags, spellId, spellName)
	local learnt = false
	if player then
		local class = nil
		local level = nil
		local race = nil
		local isEnemy = true
		local isGuess = true

		local playerData = SpyPerCharDB.PlayerData[player]
		if not playerData or playerData.isEnemy == nil then
			learnt = true
		end

		if analyseSpell then
			local type = strsub(event, 1, 5)
			if type == "SWING" or type == "SPELL" or type == "RANGE" then
				local ability = Spy_AbilityList[spellName]
				if ability then
					if ability.class and not (playerData and playerData.class) then
						class = ability.class
						learnt = true
					end
					if ability.level and (not playerData or not playerData.level or playerData.level < ability.level) then
						level = ability.level
						learnt = true
					end
					if ability.race and not (playerData and playerData.race) then
						race = ability.race
						learnt = true
					end
				end
				if class and race and level == Spy.MaximumPlayerLevel then
					isGuess = false
					learnt = true
				end
			end
		end

		Spy:UpdatePlayerData(player, class, level, race, nil, isEnemy, isGuess)
		return learnt, playerData
	end
	return learnt, nil
end

function Spy:ParseUnitDetails(player, class, level, race, guild, zone, subZone, mapX, mapY)
	if player then
		local playerData = SpyPerCharDB.PlayerData[player]
		if not playerData then
			playerData = Spy:AddPlayerData(player, class, level, race, guild, true, true)
		else
			if not playerData.class then playerData.class = class end
			if not playerData.level or playerData.level < level then playerData.level = level end
			if not playerData.race then playerData.race = race end
			if not playerData.guild then playerData.guild = guild end
		end
		playerData.isEnemy = true
		playerData.time = time()
		playerData.zone = zone
		playerData.subZone = subZone
		playerData.mapX = mapX
		playerData.mapY = mapY

		return true, playerData
	end
	return true, nil
end

function Spy:AddDetected(player, timestamp, learnt, source)
	if not Spy.NearbyList[player] then
		if Spy.db.profile.ShowOnDetection and not Spy.db.profile.MainWindowVis then
			Spy:SetCurrentList(1)
			Spy:EnableSpy(true, true)
		end
		if Spy.db.profile.CurrentList ~= 1 and Spy.db.profile.MainWindowVis and Spy.db.profile.ShowNearbyList then
			Spy:SetCurrentList(1)
		end

		if source and source ~= UnitName("player") and not Spy.ActiveList[player] then
			Spy.NearbyList[player] = timestamp
			Spy.LastHourList[player] = timestamp
			Spy.InactiveList[player] = timestamp
		else
			Spy.NearbyList[player] = timestamp
			Spy.LastHourList[player] = timestamp
			Spy.ActiveList[player] = timestamp
			Spy.InactiveList[player] = nil
		end

		if Spy.db.profile.CurrentList == 1 then
			Spy:RefreshCurrentList(player, source)
		else
			if not source or source ~= UnitName("player") then
				Spy:AlertPlayer(player)
				if not source then Spy:AnnouncePlayer(player) end
			end
		end
	elseif not Spy.ActiveList[player] then
		if Spy.db.profile.ShowOnDetection and not Spy.db.profile.MainWindowVis then
			Spy:SetCurrentList(1)
			Spy:EnableSpy(true, true)
		end
		if Spy.db.profile.CurrentList ~= 1 and Spy.db.profile.MainWindowVis and Spy.db.profile.ShowNearbyList then
			Spy:SetCurrentList(1)
		end

		Spy.LastHourList[player] = timestamp
		Spy.ActiveList[player] = timestamp
		Spy.InactiveList[player] = nil

		if Spy.db.profile.CurrentList == 1 then
			Spy:RefreshCurrentList()
		end
	else
		Spy.ActiveList[player] = timestamp
		if learnt and Spy.db.profile.CurrentList == 1 then
			Spy:RefreshCurrentList()
		end
	end
end

Spy.ListTypes = {
	{L["Nearby"], Spy.ManageNearbyList, Spy.ManageNearbyListExpirations},
	{L["LastHour"], Spy.ManageLastHourList, Spy.ManageLastHourListExpirations},
	{L["Ignore"], Spy.ManageIgnoreList},
	{L["KillOnSight"], Spy.ManageKillOnSightList},
}
